from typing import Optional


class CustomException(Exception):
    """
    Base class for custom exceptions
    """

    def __init__(
        self,
        message: Optional[str] = None,
        extra_info: Optional[dict] = None,
    ):
        self.message = message
        self.extra_info = extra_info
        super().__init__(self.message)

    def __str__(self):
        if self.extra_info:
            return f"{self.message} (Extra Info: {self.extra_info})"
        return self.message



import os
import aioshutil as async_shutil
import tempfile
import warnings
from typing import List, Optional, Union, Iterable
from datetime import datetime
import aiofiles
import aiofiles.os as async_os
import asyncio

# Package Imports
from ..processor import (
    convert_pdf_to_images,
    download_file,
    process_page,
    process_pages_in_batches,
    create_selected_pages_pdf,
)
from ..errors import FileUnavailable
from ..constants.messages import Messages
from ..models import litellmmodel
from .types import Page, ZeroxOutput


async def zerox(
    cleanup: bool = True,
    concurrency: int = 10,
    file_path: Optional[str] = "",
    maintain_format: bool = False,
    model: str = "gpt-4o-mini",
    output_dir: Optional[str] = None,
    temp_dir: Optional[str] = None,
    custom_system_prompt: Optional[str] = None,
    select_pages: Optional[Union[int, Iterable[int]]] = None,
    **kwargs
) -> ZeroxOutput:
    """
    API to perform OCR to markdown using Vision models.
    Please setup the environment variables for the model and model provider before using this API. Refer: https://docs.litellm.ai/docs/providers

    :param cleanup: Whether to cleanup the temporary files after processing, defaults to True
    :type cleanup: bool, optional
    :param concurrency: The number of concurrent processes to run, defaults to 10
    :type concurrency: int, optional
    :param file_path: The path or URL to the PDF file to process.
    :type file_path: str, optional
    :param maintain_format: Whether to maintain the format from the previous page, defaults to False
    :type maintain_format: bool, optional
    :param model: The model to use for generating completions, defaults to "gpt-4o-mini". Note - Refer: https://docs.litellm.ai/docs/providers to pass correct model name as according to provider it might be different from actual name.
    :type model: str, optional
    :param output_dir: The directory to save the markdown output, defaults to None
    :type output_dir: str, optional
    :param temp_dir: The directory to store temporary files, defaults to some named folder in system's temp directory. If already exists, the contents will be deleted for zerox uses it.
    :type temp_dir: str, optional
    :param custom_system_prompt: The system prompt to use for the model, this overrides the default system prompt of zerox. Generally it is not required unless you want some specific behaviour. When set, it will raise a friendly warning, defaults to None
    :type custom_system_prompt: str, optional
    :param select_pages: Pages to process, can be a single page number or an iterable of page numbers, defaults to None
    :type select_pages: int or Iterable[int], optional

    :param kwargs: Additional keyword arguments to pass to the model.completion -> litellm.completion method. Refer: https://docs.litellm.ai/docs/providers and https://docs.litellm.ai/docs/completion/input
    :return: The markdown content generated by the model.
    """


    input_token_count = 0
    output_token_count = 0
    prior_page = ""
    aggregated_markdown: List[str] = []
    start_time = datetime.now()
    
    # File Path Validators
    if not file_path:
        raise FileUnavailable()
    
    # Create an instance of the litellm model interface
    vision_model = litellmmodel(model=model,**kwargs)

    # override the system prompt if a custom prompt is provided
    if custom_system_prompt:
        vision_model.system_prompt = custom_system_prompt

    # Check if both maintain_format and select_pages are provided
    if maintain_format and select_pages is not None:
        warnings.warn(Messages.MAINTAIN_FORMAT_SELECTED_PAGES_WARNING)

    # If select_pages is a single integer, convert it to a list for consistency
    if isinstance(select_pages, int):
        select_pages = [select_pages]

    # Sort the pages to maintain consistency
    if select_pages is not None:
        select_pages = sorted(select_pages)

    # Ensure the output directory exists
    if output_dir:
        await async_os.makedirs(output_dir, exist_ok=True)

    ## delete tmp_dir if exists and then recreate it
    if temp_dir:
        if os.path.exists(temp_dir):
            await async_shutil.rmtree(temp_dir)
        await async_os.makedirs(temp_dir, exist_ok=True)


    # Create a temporary directory to store the PDF and images
    with tempfile.TemporaryDirectory() as temp_dir_:

        if temp_dir:
            ## use the user provided temp directory
            temp_directory = temp_dir
        else:
            ## use the system temp directory
            temp_directory = temp_dir_

        # Download the PDF. Get file name.
        local_path = await download_file(file_path=file_path, temp_dir=temp_directory)
        if not local_path:
            raise FileUnavailable()
        
        raw_file_name = os.path.splitext(os.path.basename(local_path))[0]
        file_name = "".join(c.lower() if c.isalnum() else "_" for c in raw_file_name)
        # Truncate file name to 255 characters to prevent ENAMETOOLONG errors
        file_name = file_name[:255]

        # create a subset pdf in temp dir with only the requested pages if select_pages is provided
        if select_pages is not None:
            subset_pdf_create_kwargs = {"original_pdf_path":local_path, "select_pages":select_pages, 
                                    "save_directory":temp_directory, "suffix":"_selected_pages"}
            local_path = await asyncio.to_thread(create_selected_pages_pdf, 
                                                 **subset_pdf_create_kwargs)

        # Convert the file to a series of images, below function returns a list of image paths in page order
        images = await convert_pdf_to_images(local_path=local_path, temp_dir=temp_directory)

        if maintain_format:
            for image in images:
                result, input_token_count, output_token_count, prior_page = await process_page(
                    image,
                    vision_model,
                    temp_directory,
                    input_token_count,
                    output_token_count,
                    prior_page,
                )

                if result:
                    aggregated_markdown.append(result)
        else:
            results = await process_pages_in_batches(
                images,
                concurrency,
                vision_model,
                temp_directory,
                input_token_count,
                output_token_count,
                prior_page,
            )

            aggregated_markdown = [result[0] for result in results if isinstance(result[0], str)]

            ## add token usage
            input_token_count += sum([result[1] for result in results])
            output_token_count += sum([result[2] for result in results])

        # Write the aggregated markdown to a file
        if output_dir:
            result_file_path = os.path.join(output_dir, f"{file_name}.md")
            async with aiofiles.open(result_file_path, "w", encoding="utf-8") as f:
                await f.write("\n\n".join(aggregated_markdown))

        # Cleanup the downloaded PDF file
        if cleanup and os.path.exists(temp_directory):
            await async_shutil.rmtree(temp_directory)

        # Format JSON response
        end_time = datetime.now()
        completion_time = (end_time - start_time).total_seconds() * 1000

        # Adjusting the formatted_pages logic to account for select_pages to output the correct page numbers
        if select_pages is not None:
            # Map aggregated markdown to the selected pages
            formatted_pages = [
                        Page(content=content, page=select_pages[i], content_length=len(content))
                        for i, content in enumerate(aggregated_markdown)
                    ]
        else:
            # Default behavior when no select_pages is provided
            formatted_pages = [
                        Page(content=content, page=i + 1, content_length=len(content))
                        for i, content in enumerate(aggregated_markdown)
                    ]

        return ZeroxOutput(
            completion_time=completion_time,
            file_name=file_name,
            input_tokens=input_token_count,
            output_tokens=output_token_count,
            pages=formatted_pages,
        )


import os
import aiohttp
import warnings
import litellm
from typing import List, Dict, Any, Optional

# Package Imports
from .base import BaseModel
from .types import CompletionResponse
from ..errors import ModelAccessError, NotAVisionModel, MissingEnvironmentVariables
from ..constants.messages import Messages
from ..constants.prompts import Prompts
from ..processor.image import encode_image_to_base64

DEFAULT_SYSTEM_PROMPT = Prompts.DEFAULT_SYSTEM_PROMPT


class litellmmodel(BaseModel):
    ## setting the default system prompt
    _system_prompt = DEFAULT_SYSTEM_PROMPT

    def __init__(
        self,
        model: Optional[str] = None,
        **kwargs,
    ):
        """
        Initializes the Litellm model interface.
        :param model: The model to use for generating completions, defaults to "gpt-4o-mini". Refer: https://docs.litellm.ai/docs/providers
        :type model: str, optional
        
        :param kwargs: Additional keyword arguments to pass to self.completion -> litellm.completion. Refer: https://docs.litellm.ai/docs/providers and https://docs.litellm.ai/docs/completion/input
        """
        super().__init__(model=model, **kwargs)

        ## calling custom methods to validate the environment and model
        self.validate_environment()
        self.validate_model()
        self.validate_access()

    @property
    def system_prompt(self) -> str:
        '''Returns the system prompt for the model.'''
        return self._system_prompt
    
    @system_prompt.setter
    def system_prompt(self, prompt: str) -> None:
        '''
        Sets/overrides the system prompt for the model.
        Will raise a friendly warning to notify the user. 
        '''
        warnings.warn(f"{Messages.CUSTOM_SYSTEM_PROMPT_WARNING}. Default prompt for zerox is:\n {DEFAULT_SYSTEM_PROMPT}")
        self._system_prompt = prompt

    ## custom method on top of BaseModel
    def validate_environment(self) -> None:
        """Validates the environment variables required for the model."""
        env_config = litellm.validate_environment(model=self.model)

        if not env_config["keys_in_environment"]:
            raise MissingEnvironmentVariables(extra_info=env_config)
        
    def validate_model(self) -> None:
        '''Validates the model to ensure it is a vision model.'''
        if not litellm.supports_vision(model=self.model):
            raise NotAVisionModel(extra_info={"model": self.model})
        
    def validate_access(self) -> None:
        """Validates access to the model -> if environment variables are set correctly with correct values."""
        if not litellm.check_valid_key(model=self.model,api_key=None):
            raise ModelAccessError(extra_info={"model": self.model})
        

    async def completion(
        self,
        image_path: str,
        maintain_format: bool,
        prior_page: str,
    ) -> CompletionResponse:
        """LitellM completion for image to markdown conversion.

        :param image_path: Path to the image file.
        :type image_path: str
        :param maintain_format: Whether to maintain the format from the previous page.
        :type maintain_format: bool
        :param prior_page: The markdown content of the previous page.
        :type prior_page: str

        :return: The markdown content generated by the model.
        """
        messages = await self._prepare_messages(
            image_path=image_path,
            maintain_format=maintain_format,
            prior_page=prior_page,
        )

        try:
            response = await litellm.acompletion(model=self.model, messages=messages, **self.kwargs)

            ## completion response
            response = CompletionResponse(
                    content=response["choices"][0]["message"]["content"],
                    input_tokens=response["usage"]["prompt_tokens"],
                    output_tokens=response["usage"]["completion_tokens"],
                )
            return response
        
        except Exception as err:
            raise Exception(Messages.COMPLETION_ERROR.format(err))

    async def _prepare_messages(
        self,
        image_path: str,
        maintain_format: bool,
        prior_page: str,
    ) -> List[Dict[str, Any]]:
        """Prepares the messages to send to the LiteLLM Completion API.

        :param image_path: Path to the image file.
        :type image_path: str
        :param maintain_format: Whether to maintain the format from the previous page.
        :type maintain_format: bool
        :param prior_page: The markdown content of the previous page.
        :type prior_page: str
        """
        # Default system message
        messages: List[Dict[str, Any]] = [
            {
                "role": "system",
                "content": self._system_prompt,
            },
        ]

        # If content has already been generated, add it to context.
        # This helps maintain the same format across pages.
        if maintain_format and prior_page:
            messages.append(
                {
                    "role": "system",
                    "content": f'Markdown must maintain consistent formatting with the following page: \n\n """{prior_page}"""',
                },
            )

        # Add Image to request
        base64_image = await encode_image_to_base64(image_path)
        messages.append(
            {
                "role": "user",
                "content": [
                    {
                        "type": "image_url",
                        "image_url": {"url": f"data:image/png;base64,{base64_image}"},
                    },
                ],
            }
        )

        return messages



import re

# Package imports
from ..constants.patterns import Patterns


def format_markdown(text: str) -> str:
    """Format markdown text by removing markdown and code blocks"""

    formatted_markdown = re.sub(Patterns.MATCH_MARKDOWN_BLOCKS, r"\1", text)
    formatted_markdown = re.sub(Patterns.MATCH_CODE_BLOCKS, r"\1", formatted_markdown)
    return formatted_markdown



import aiofiles
import base64
import io


async def encode_image_to_base64(image_path: str) -> str:
    """Encode an image to base64 asynchronously."""
    async with aiofiles.open(image_path, "rb") as image_file:
        image_data = await image_file.read()
    return base64.b64encode(image_data).decode("utf-8")


async def save_image(image, image_path: str):
    """Save an image to a file asynchronously."""
    # Convert PIL Image to BytesIO object
    with io.BytesIO() as buffer:
        image.save(buffer, format=image.format)  # Save the image to the BytesIO object
        image_data = buffer.getvalue()  # Get the image data from the BytesIO object

    # Write image data to file asynchronously
    async with aiofiles.open(image_path, "wb") as f:
        await f.write(image_data)



from .image import save_image, encode_image_to_base64
from .pdf import (
    convert_pdf_to_images,
    process_page,
    process_pages_in_batches,
)
from .text import format_markdown
from .utils import download_file, create_selected_pages_pdf

__all__ = [
    "save_image",
    "encode_image_to_base64",
    "convert_pdf_to_images",
    "format_markdown",
    "download_file",
    "process_page",
    "process_pages_in_batches",
    "create_selected_pages_pdf",
]



from typing import List, Optional, Dict, Any, Union, Iterable
from dataclasses import dataclass, field


@dataclass
class ZeroxArgs:
    """
    Dataclass to store the arguments for the Zerox class.
    """

    file_path: str
    cleanup: bool = True
    concurrency: int = 10
    maintain_format: bool = False
    model: str = "gpt-4o-mini",
    output_dir: Optional[str] = None
    temp_dir: Optional[str] = None
    custom_system_prompt: Optional[str] = None
    select_pages: Optional[Union[int, Iterable[int]]] = None
    kwargs: Dict[str, Any] = field(default_factory=dict)

@dataclass
class Page:
    """
    Dataclass to store the page content.
    """

    content: str
    content_length: int
    page: int


@dataclass
class ZeroxOutput:
    """
    Dataclass to store the output of the Zerox class.
    """

    completion_time: float
    file_name: str
    input_tokens: int
    output_tokens: int
    pages: List[Page]



from dataclasses import dataclass


@dataclass
class CompletionResponse:
    """
    A class representing the response of a completion.
    """

    content: str
    input_tokens: int
    output_tokens: int



import os
import re
from typing import Optional, Union, Iterable
from urllib.parse import urlparse
import aiofiles
import aiohttp
from PyPDF2 import PdfReader, PdfWriter
from ..constants.messages import Messages

# Package Imports
from ..errors.exceptions import ResourceUnreachableException, PageNumberOutOfBoundError


async def download_file(
    file_path: str,
    temp_dir: str,
) -> Optional[str]:
    """Downloads a file from a URL or local path to a temporary directory."""

    local_pdf_path = os.path.join(temp_dir, os.path.basename(file_path))
    if is_valid_url(file_path):
        async with aiohttp.ClientSession() as session:
            async with session.get(file_path) as response:
                if response.status != 200:
                    raise ResourceUnreachableException()
                async with aiofiles.open(local_pdf_path, "wb") as f:
                    await f.write(await response.read())
    else:
        async with aiofiles.open(file_path, "rb") as src, aiofiles.open(
            local_pdf_path, "wb"
        ) as dst:
            await dst.write(await src.read())
    return local_pdf_path


def is_valid_url(string: str) -> bool:
    """Checks if a string is a valid URL."""

    try:
        result = urlparse(string)
        return all([result.scheme, result.netloc]) and result.scheme in [
            "http",
            "https",
        ]
    except ValueError:
        return False
    
def create_selected_pages_pdf(original_pdf_path: str, select_pages: Union[int, Iterable[int]], 
                              save_directory: str, suffix: str = "_selected_pages",
                              sorted_pages: bool = True) -> str:
    """
    Creates a new PDF with only the selected pages.
    
    :param original_pdf_path: Path to the original PDF file.
    :type original_pdf_path: str
    :param select_pages: A single page number or an iterable of page numbers (1-indexed).
    :type select_pages: int or Iterable[int]
    :param save_directory: The directory to store the new PDF.
    :type save_directory: str
    :param suffix: The suffix to add to the new PDF file name, defaults to "_selected_pages".
    :type suffix: str, optional
    :param sorted_pages: Whether to sort the selected pages, defaults to True.
    :type sorted_pages: bool, optional
    :return: Path the new PDF file
    """

    file_name = os.path.splitext(os.path.basename(original_pdf_path))[0]

    # Write the new PDF to a temporary file
    selected_pages_pdf_path = os.path.join(save_directory, f"{file_name}{suffix}.pdf")

    # Ensure select_pages is iterable, if not, convert to list
    if isinstance(select_pages, int):
        select_pages = [select_pages]
    
    if sorted_pages:
        # Sort the pages for consistency
        select_pages = sorted(list(select_pages))

    with open(original_pdf_path, "rb") as orig_pdf, open(selected_pages_pdf_path, "wb") as new_pdf:

        # Read the original PDF
        reader = PdfReader(stream=orig_pdf)
        total_pages = len(reader.pages)

        # Validate page numbers
        invalid_page_numbers = []
        for page in select_pages:
            if page < 1 or page > total_pages:
                invalid_page_numbers.append(page)

        ## raise error if invalid page numbers
        if invalid_page_numbers:
            raise PageNumberOutOfBoundError(extra_info={"input_pdf_num_pages":total_pages,
                                                        "select_pages": select_pages,
                                                        "invalid_page_numbers": invalid_page_numbers})

        # Create a new PDF writer
        writer = PdfWriter(fileobj=new_pdf)

        # Add only the selected pages
        for page_number in select_pages:
            writer.add_page(reader.pages[page_number - 1])

        writer.write(stream=new_pdf)

    return selected_pages_pdf_path



from abc import ABC, abstractmethod
from typing import Dict, Optional, Type, TypeVar, TYPE_CHECKING

if TYPE_CHECKING:
    from ..models import CompletionResponse

T = TypeVar("T", bound="BaseModel")

class BaseModel(ABC):
    """
    Base class for all models.
    """

    @abstractmethod
    async def completion(
        self,
    ) -> "CompletionResponse":
        raise NotImplementedError("Subclasses must implement this method")
    
    @abstractmethod
    def validate_access(
        self,
    ) -> None:
        raise NotImplementedError("Subclasses must implement this method")
    
    @abstractmethod
    def validate_model(
        self,
    ) -> None:
        raise NotImplementedError("Subclasses must implement this method")

    def __init__(
        self,
        model: Optional[str] = None,
        **kwargs,
    ):
        self.model = model
        self.kwargs = kwargs

        ## validations
        # self.validate_model()
        # self.validate_access()





class Prompts:
    """Class for storing prompts for the Zerox system."""

    DEFAULT_SYSTEM_PROMPT = """
    Convert the following PDF page to markdown.
    Return only the markdown with no explanation text.
    Do not exclude any content from the page.
    If there is more than one document in the same PDF
    or image, separate the Markdown for each document as
    separate sections.
    """


import logging
import os
import asyncio
from typing import List, Optional, Tuple
from pdf2image import convert_from_path

# Package Imports
from .image import save_image
from .text import format_markdown
from ..constants import PDFConversionDefaultOptions, Messages
from ..models import litellmmodel


async def convert_pdf_to_images(local_path: str, temp_dir: str) -> List[str]:
    """Converts a PDF file to a series of images in the temp_dir. Returns a list of image paths in page order."""
    options = {
        "pdf_path": local_path,
        "output_folder": temp_dir,
        "dpi": PDFConversionDefaultOptions.DPI,
        "fmt": PDFConversionDefaultOptions.FORMAT,
        "size": PDFConversionDefaultOptions.SIZE,
        "thread_count": PDFConversionDefaultOptions.THREAD_COUNT,
        "use_pdftocairo": PDFConversionDefaultOptions.USE_PDFTOCAIRO,
        "paths_only": True,
    }

    try:
        image_paths = await asyncio.to_thread(
            convert_from_path, **options
        )
        return image_paths
    except Exception as err:
        logging.error(f"Error converting PDF to images: {err}")


async def process_page(
    image: str,
    model: litellmmodel,
    temp_directory: str = "",
    input_token_count: int = 0,
    output_token_count: int = 0,
    prior_page: str = "",
    semaphore: Optional[asyncio.Semaphore] = None,
) -> Tuple[str, int, int, str]:
    """Process a single page of a PDF"""

    # If semaphore is provided, acquire it before processing the page
    if semaphore:
        async with semaphore:
            return await process_page(
                image,
                model,
                temp_directory,
                input_token_count,
                output_token_count,
                prior_page,
            )

    image_path = os.path.join(temp_directory, image)

    # Get the completion from LiteLLM
    try:
        completion = await model.completion(
            image_path=image_path,
            maintain_format=True,
            prior_page=prior_page,
        )

        formatted_markdown = format_markdown(completion.content)
        input_token_count += completion.input_tokens
        output_token_count += completion.output_tokens
        prior_page = formatted_markdown

        return formatted_markdown, input_token_count, output_token_count, prior_page

    except Exception as error:
        logging.error(f"{Messages.FAILED_TO_PROCESS_IMAGE} Error:{error}")
        return "", input_token_count, output_token_count, ""


async def process_pages_in_batches(
    images: List[str],
    concurrency: int,
    model: litellmmodel,
    temp_directory: str = "",
    input_token_count: int = 0,
    output_token_count: int = 0,
    prior_page: str = "",
):
    # Create a semaphore to limit the number of concurrent tasks
    semaphore = asyncio.Semaphore(concurrency)

    # Process each page in parallel
    tasks = [
        process_page(
            image,
            model,
            temp_directory,
            input_token_count,
            output_token_count,
            prior_page,
            semaphore,
        )
        for image in images
    ]

    # Wait for all tasks to complete
    return await asyncio.gather(*tasks)
